import binascii

class Forwarder( Thread ):

    START_PORT = 10000
    _currentport = START_PORT
    

    def __init__( self, sessionID, addr, ports,
                  addlock, dellock, playlock, pauselock, newserverlock, printflag ):    
        Thread.__init__( self )

        self._addlock = addlock
        self._dellock = dellock
        self._playlock = playlock
        self._pauselock = pauselock
        self._newserverlock = newserverlock
        
        # Init structure for communication with parent Session thread
        self._addsessionid = []
        self._addaddr = []
        self._addports = []
        self._delsessionid = []
        self._playsessionid = []
        self._pausesessionid = []
        self._newserveraddr = []
        self._newserverports = []

        self._portrange = None
        self._printflag = printflag
        
        # for gathering statistics
        self._stats = {}
        self._clients_paused = { sessionID: (gethostbyname( addr ), ports) }
        self._clients = {}
        self.buildports( sessionID )
        
	self.setDaemon( 1 )

    def __del__(self):
        for k in self._proxysocks:
            self._proxysocks[k][1].close()
            
    def addclient( self, sessionID, addr, ports ):
        self._clients_paused[sessionID] = ( gethostbyname( addr ), ports )

    def delclient( self, sessionID ):
        if self._clients.has_key(sessionID):
            del self._clients[sessionID]
        if self._clients_paused.has_key(sessionID):
            del self._clients_paused[sessionID]


    def makeclientactive( self, sessionID ):
        if self._clients_paused.has_key(sessionID):
            self._clients[sessionID] = self._clients_paused[sessionID]
            del self._clients_paused[sessionID]
        else:
            debug("FORWARDER: Illegal input to Forwarder.makeclientactive. " +
                  "SessionID %s not present.\n" % sessionID)

    def makeclientpaused( self, sessionID ):
        if self._clients.has_key(sessionID):
            self._clients_paused[sessionID] = self._clients[sessionID]
            del self._clients[sessionID]
        else:
            debug("FORWARDER: Illegal input to Forwarder.makeclientpaused. " +
                  "SessionID %s not present.\n" % sessionID)

        
    def doforwarding( self ):    
        sockets = map( lambda x: x[1], self._proxysocks )
        count = 0
        size = 0
        seqno = 0
        then = time.time()    
        while 1:
            #debug("FORWARDER: In loop")
            readylist, _, _ = select( sockets, [], [] )            
            for i in range(len(self._proxysocks)):            
                port, sock = self._proxysocks[i]                
                if sock in readylist:
                    (packet, addr) = sock.recvfrom( 2048 )
                    serverport = self._serverports[i]
                    if addr == (self._serveraddr, serverport):
                        count = count + 1
                        size = size + len(packet)
                        seqno = seqno + 1

                        keys = self._clients.keys()
                        for j in range(len(keys)):
                            sock.sendto( packet, (self._clients[keys[j]][0],
                                                  self._clients[keys[j]][1][i]) )
      
            # Update active and paused clients
            # Add
            if self._addlock.acquire(0):
                if len(self._addsessionid) > 0:
                    sessionID = self._addsessionid.pop(0)
                    addr = self._addaddr.pop(0)
                    ports = self._addports.pop(0)
                    self.addclient(sessionID, addr, ports)
                    debug("FORWARDER: Added client %s" % sessionID)
                self._addlock.release()
            # Delete
            if self._dellock.acquire(0):
                if len(self._delsessionid) > 0:
                    sessionID = self._delsessionid.pop(0)
                    self.delclient(sessionID)
                    debug("FORWARDER: Deleted client %s" % sessionID)
                self._dellock.release()
            # Make active
            if self._playlock.acquire(0):
                if len(self._playsessionid) > 0:
                    sessionID = self._playsessionid.pop(0)
                    self.makeclientactive(sessionID)
                    debug("FORWARDER: Made active client %s" % sessionID)
                self._playlock.release()
            # Make inactive
            if self._pauselock.acquire(0):
                if len(self._pausesessionid) > 0:
                    sessionID = self._pausesessionid.pop(0)
                    self.makeclientpaused(sessionID)
                    debug("FORWARDER: Made paused client %s" % sessionID)
                self._pauselock.release()                
            # See if we need to be redirected
            if self._newserverlock.acquire(0):
                if len(self._newserveraddr) > 0:
		    debug("FORWARDER: FOUND new parent")
                    addr = self._newserveraddr.pop(0)
                    ports = self._newserverports.pop(0)
                    self.setserver(addr, ports)
                self._newserverlock.release()

    def run( self ):    
        debug( "FORWARDER:  Starting an RTP forwarder on ports " + self.getportrange() )
        debug( "    server %s:%s  <->  client %s:%s\n" % (
            self._serveraddr, makeportrange(self._serverports),
            self._clients_paused[self._clients_paused.keys()[0]][0],
            makeportrange(self._clients_paused[self._clients_paused.keys()[0]][1])))
        self.doforwarding()
    
    def buildports( self, sessionID ):
    
        num = len( self._clients_paused[sessionID][1] )
        self._proxysocks = self._allocateports( num )
        self._portrange = self.getportrange()
        
    def _allocateports( self, howmany ):    
        start = Forwarder._currentport
        sofar = 0
        socks = []
        
        while sofar < howmany:
        
            sock = socket( AF_INET, SOCK_DGRAM )
            port = Forwarder._currentport
            Forwarder._currentport = Forwarder._currentport + 1
            
            try:
                sock.bind( ('',port) )
                socks.append( (port,sock) )
                sofar = sofar + 1
                end = port
            except:
                sofar = 0
                if Forwarder._currentport % 2 == 1:
                    Forwarder._currentport = Forwarder._currentport + 1
                start = Forwarder._currentport
                socks = []
                
        debug( "FORWARDER: allocated a port range at %d-%d\n" % (start,end) )

        return socks


    def setserver( self, addr, ports ):
    
        self._serveraddr = gethostbyname( addr )
        self._serverports = ports
    
    def getportrange( self ):
    
        ports = map( lambda x: x[0], self._proxysocks )
        return makeportrange( ports )


