class Session( Thread ):

    RTSP_PORT = 554
    MAX_CHILD = 2
    DEFAULT_PORT = 5982
    DUMMY_SESSIONID = "0123456789"
    
    def __init__( self, url, addlock, dellock, manager, port, target, killlock ):
        debug("SESSION: Initializing Session")
        Thread.__init__( self )

        # Store locks and parent info
        self._url = url
        self._addlock = addlock
        self._dellock = dellock
        self._manager = manager

        self._newclientconn = []
        self._newclientaddr = []
        self._newclientport = []
        self._newclienttype = []
        self._newclientmsg = []

        # Init server info
        self._serverconn = None
        self._serveraddr = None
	self._serverport = None
        self._serversessionid = ''

        # Make a PRNG
        self._prng = random.Random()

        self._trackIDs = []
        self._tracknum = 0
        
        # Store client info
        self._client_conns = {}
        self._client_addrs = {}
        self._client_ports = {}
        self._client_types = {}
        self._forwarders = {}
        
        # Init replies
        self._DESCRIBEreply = None
        self._SETUPreplies = {}
        self._PLAYreply = None
        self._PAUSEreply = None
        self._TEARDOWNreply = None

        self._DESCRIBErequest = None
        self._SETUPrequests = {}
        self._PLAYrequest = None
        self._PAUSErequest = None
        self._TEARDOWNrequest = None
        
        # To enable proxy<->server communication
        self._proxysessionID = None
        self._proxycseqval = 0
	self.setDaemon( 1 )

        self._amDead = 0
        self._die = 0
        self._nextChild = None
        self._localPort = port
        self._target = target
        self._killlock = killlock

    def addclient( self, conn, addr, cport, type ):
        debug("SESSION: Adding client")
        sessionID = self.getsessionID()
        self._client_conns[sessionID] = conn
        self._client_addrs[sessionID] = addr
        self._client_ports[sessionID] = cport
        self._client_types[sessionID] = type
        return sessionID

    def delclient( self, sessionID ):
        debug("SESSION: Removing client")
        if self._client_conns.has_key(sessionID):
            self._client_conns[sessionID].close()
            del self._client_conns[sessionID]
        if self._client_addrs.has_key(sessionID):
            del self._client_addrs[sessionID]
        if self._client_ports.has_key(sessionID):
            del self._client_ports[sessionID]
        if self._client_types.has_key(sessionID):
            del self._client_types[sessionID]
            
#----------------------------- Utilities -----------------------------------------------------------        
    def getsessionID( self ):
        c = str(self._prng.random())
	return c[2:]
        
    def getclientmsg( self, sessionID ):
        if self._client_conns.has_key(sessionID):
            clientconn = self._client_conns[sessionID]
            return Message( clientconn )
        else:
            debug("SESSION: Illegal input to Session.getclientmsg. SessionID %s does not exist\n" % sessionID)
            return None

    def sendclientmsg( self, sessionID, msg ):
        if self._client_conns.has_key(sessionID):
            clientconn = self._client_conns[sessionID]
            clientconn.send( msg.getmessage() )
        else:
            debug("SESSION: Illegal input to Session.sendclientmsg. SessionID %s does not exist\n" % sessionID)
            return None

    
    def getservermsg( self ):
        return Message( self._serverconn )

    def sendservermsg( self, msg ):
        self._serverconn.send( msg.getmessage() )
#----------------------------  DISPATCH  -----------------------------------------------
    
    def dispatch( self, sessionID, msg ):
        debug("CLIENT -> PROXY:\n")
        debug(msg.getmessage())
        command = msg.getcommand()
        
        if command == "DESCRIBE":
            self.do_describe( sessionID, msg )            
        elif command == "SETUP":
            self.do_setup( sessionID, msg )
        elif command == "PLAY":
            self.do_play( sessionID, msg )
        elif command == "PAUSE":
            self.do_pause( sessionID, msg )
        elif command == "TEARDOWN":
            self.do_teardown( sessionID, msg)
        else:
            debug("SESSION: Illegal input to Session.dispatch\n%s\n" % msg.getmessage())
            
#        elif command == "OPTIONS":
#            self.do_options( sessionID, msg )
#            
#        else:
#            self.sendservermsg( msg )
#            response = self.getservermsg()
#            self.sendclientmsg( sessionID, response )

    
#    def do_options( self, msg ):
#        if self._client_type is not None:
#            if self._client_type[:4] == 'QTS/' and self._server_type == 'QTSS/v66':
#                debug( '  translating OPTIONS into a GET_PARAMETER ping for broken QuickTime' )
#                msg.setcommand( 'GET_PARAMETER' )
#                self.sendservermsg( msg )
#                msg.setcommand( 'RTSP/1.0' )
#                msg.setargs( ['200', 'OK'] )
#                self.sendclientmsg( msg )
#        else:
#            self.sendservermsg( msg )
#            response = self.getservermsg()
#            self.sendclientmsg( response )


#-------------------------- DESCRIBE   ----------------------------------------------

    def do_describe( self, sessionID, msg ):
        # Get user agent info
        if sessionID != self.DUMMY_SESSIONID:
            self._client_types[sessionID] = msg.getheader('user-agent')

            if len(self._client_conns) > self.MAX_CHILD:
                self.do_redirect(sessionID, msg)
                return
                
        if not self._serverconn:
            # Make the RTSP request sane
            args = msg.getargs()
            url = args[0]
            parsed_url = urlparse.urlparse( url, "rtsp" )
            target = parsed_url[4]
            newargs = args[:]
            newargs[0] = target
            msg.setargs(newargs)
            self._proxycseqval = self._proxycseqval + 1
            msg.setheader('cseq', self._proxycseqval)

            # Change the user-agent and port information in request
            if msg._headerdict.has_key('user-agent'):
                msg.setheader('user-agent', 'chaincast ver 0.0')
            else:
                msg.addheader('user-agent', 'chaincast ver 0.0')
            if msg._headerdict.has_key('port'):
                msg.setheader('port', "%s" %self._localPort)
            else:
                msg.addheader('port', "%s" %self._localPort)

            # Get server info
            url = msg.getargs()[0]
            parsed_url = urlparse.urlparse( url, "rtsp" )
            site = parsed_url[1]
            debug(site)
            pos = string.find( site, ":" )
            if pos >= 0:
                addr,port = string.split( site, ":", 1)
                port = int( port )
            else:
                addr = site
                port = self.RTSP_PORT
            # debug ("here we are" + addr)
            # Connect to server
            sock = socket( AF_INET, SOCK_STREAM )
            sock.connect( (addr,port) )
            self._serverconn = sock
            self._serveraddr = addr
	    self._serverport = port

            # Send request, pass on response
            debug( "PROXY -> SERVER:\n" )
            debug( msg.getmessage() )

            # Store the DESCRIBE request - not sure if this is the best place to do it
            if self._DESCRIBErequest is None:
                self._DESCRIBErequest = Message(None, msg)
            
            self.sendservermsg( msg )
            response = self.getservermsg()

            debug( "SERVER -> PROXY:\n" )
            debug( response.getmessage() )

            # React to REDIRECT request, if any
            if response.getcommand() == "REDIRECT":
                newServer = response.getheader('Location')
                oldArgs = msg.getargs()
                oldArgs[0] = newServer
                msg.setargs(oldArgs)

	        print "*********** REDIRECT received ************"
		print response.getmessage()
 
                # Do cleanup
                self._serverconn.close()
                self._serverconn = None
                self._serveraddr = None
		self._serverport = None
                self._proxycseqval = self._proxycseqval - 1
                self.dispatch(sessionID, msg)
                return
            
            self._server_type = response.getheader('server')

            # Parse response and store info for future
            self._DESCRIBEreply = response
            content = response.getcontent()
            pos = string.find(content, "a=control:trackID=")
            while pos <> -1:
                endpos = string.find(content, "\n", pos+18)
                s = content[pos+18:endpos-1]
                trackID = int(s)
                self._trackIDs.append(trackID)
                self._tracknum = self._tracknum + 1
                pos = string.find(content, "a=control:trackID=", endpos+2)
                
        else:
            # Retrieve DESCRIBEreply, make substitutions and send
            response = Message(None, self._DESCRIBEreply)
            cseqVal = msg.getheader("Cseq")
            response.setheader("Cseq", cseqVal)            
            dateVal =  msg.getheader("Date")
            if dateVal is not None:
                debug ("Session: Describe: Changing date value\n" )
                dateVal = time.strftime("%a, %d %b %Y %H:%M:%S %Z", time.gmtime())
                response.setheader("Date", dateVal)                
            expiresVal =  msg.getheader("Expires")
            if expiresVal is not None:
                debug ("Session: Describe: Changing expires value\n" )
                expiresVal = time.strftime("%a, %d %b %Y %H:%M:%S %Z", time.gmtime())
                response.setheader("Expires", expiresVal)

        if sessionID != self.DUMMY_SESSIONID:
            debug( "PROXY -> CLIENT:\n")
            debug( response.getmessage() )
            self.sendclientmsg( sessionID, response )

#------------------------- SETUP     --------------------------------------------

    def do_setup( self, sessionID, msg ):
        
        # Get track ID
        url = msg.getargs()[0]
        parsed_url = urlparse.urlparse( url, "rtsp" )
        path = parsed_url[2]
        trackID = None
        
        for bit in string.split( path, "/" ):
            bit = string.strip( bit )
            if string.find( bit, "=" ) > 0:
                name, value = string.split( bit, "=", 1 )
                if string.lower(name) == "trackid":
                    trackID = int(value)
        if trackID is None:
            debug (" Session: Error TrackID is None ")
            return        

        # Get client ports
        client_port = ''        
        for bit in string.split( msg.getheader('transport'), ";" ):
            bit = string.strip( bit )            
            if string.find( bit, '=' ) > 0:
                name, value = string.split( bit, '=', 1 )            
                if name == 'client_port':
                    client_port = value            
        if string.find( client_port, "-" ):
            startport,endport = string.split( client_port, "-" )
            clientports = range( int(startport) , int(endport) + 1 )
        else:
            clientports = [ int(client_port) ]

        # Check if a forwarder exists for this track, else make a new one
        if self._forwarders.has_key(trackID):
            # If it is a normal client than do this
            if sessionID != self.DUMMY_SESSIONID:
                response = self._SETUPreplies[trackID]
                self._forwarders[trackID]._addlock.acquire()
                self._forwarders[trackID]._addsessionid.append(sessionID)
                self._forwarders[trackID]._addaddr.append(self._client_addrs[sessionID])
                self._forwarders[trackID]._addports.append(clientports)
                self._forwarders[trackID]._addlock.release()
            # If DUMMY_SESSIONID, then inform Forwarder of new server/ports
            else:
                portrange = self._forwarders[trackID]._portrange
                (response, addr, serverports) = self.do_setup_inner( msg, trackID, portrange )
                self._forwarders[trackID]._newserverlock.acquire()
                self._forwarders[trackID]._newserveraddr.append(addr)
                self._forwarders[trackID]._newserverports.append(serverports)
                self._forwarders[trackID]._newserverlock.release()
        else:
            # Make a forwarder
            addlock = Lock()
            dellock = Lock()
            playlock = Lock()
            pauselock = Lock()  
            newserverlock = Lock()
            ## TODO: what to do if DUMMY_SESSIONID occurs?

            if trackID % 2 == 0:
                printflag = 1
            else:
                printflag = 0
            proxy = Forwarder( sessionID, self._client_addrs[sessionID], clientports,
                               addlock, dellock, playlock, pauselock, newserverlock, printflag )
            self._forwarders[trackID] = proxy
            
            # Make subsitutions to msg and send to server, receive and store replies
            portrange = proxy._portrange
            (response, addr, serverports) = self.do_setup_inner( msg, trackID, portrange );
            proxy.setserver( addr, serverports )
            proxy.start()

        # Reply to the client        
        cseqVal = msg.getheader("Cseq")
        response.setheader("Cseq", cseqVal)                            
        response.setheader("Session", sessionID)
        response.setheader( 'transport',
                            'RTP/AVP;unicast;client_port=%s;server_port=%s'
                            % (client_port, self._forwarders[trackID].getportrange()) )

        if sessionID != self.DUMMY_SESSIONID:
            debug("PROXY -> CLIENT:\n")
            debug( response.getmessage() )
            self.sendclientmsg( sessionID, response )

#----------------------------- SETUP-Inner -----------------------------------------

    def do_setup_inner( self, msg, trackID, portrange ):
        msg.setheader( 'transport', 'RTP/AVP;unicast;client_port=' + portrange )
        self._proxycseqval = self._proxycseqval + 1
        msg.setheader('Cseq', self._proxycseqval)

        if self._proxysessionID is not None:
	    if msg.getheader('Session') is not None:
	        msg.setheader('Session', self._proxysessionID)
	    else:
	        msg.addheader('Session', self._proxysessionID)

	elif msg.getheader('Session') is not None:
	    msg.delheader('Session')

        # Store SETUP request for future use
        self._SETUPrequests[trackID] = Message(None, msg)

        # Send message, get response
        debug("PROXY -> SERVER:\n")
        debug( msg.getmessage() )        
        self.sendservermsg( msg )
        response = self.getservermsg()
        debug("SERVER -> PROXY:\n")
        debug( response.getmessage() )
        
        # Store response for future use
        self._SETUPreplies[trackID] = response
        self._proxysessionID = response.getheader('Session')
        server_port = ''
        source = ''        

        for bit in string.split( response.getheader('transport'), ";" ):
            bit = string.strip( bit )            
            if string.find( bit, '=' ) > 0:
                name, value = string.split( bit, '=', 1 )            
                if name == 'server_port':
                    server_port = value                
                elif name == 'source':
                    source = value        

        if string.find( server_port, "-" ):
            startport,endport = string.split( server_port, "-" )
            serverports = range( int(startport), int(endport) + 1 )
        else:
            serverports = [ int(server_port) ]
                
        if source <> '':
            addr = source
        else:
            addr = self._serveraddr

        return (response, addr, serverports)

#------------------------------- PLAY -----------------------------------------------
    def do_play( self, sessionID, msg ):
        if self._proxysessionID is not None:
            if (self._PLAYreply is None) or (sessionID == self.DUMMY_SESSIONID):
                # Make msg substitutions
                self._proxycseqval = self._proxycseqval + 1
                msg.setheader("CSeq", self._proxycseqval)
                msg.setheader("Session", self._proxysessionID)

                debug("PROXY -> SERVER:\n")
                debug( msg.getmessage() )

                # Store the PLAY request for future use
                self._PLAYrequest = Message(None, msg)


                # Send request and get response
                self.sendservermsg( msg )                
                response = self.getservermsg()

                # Store the response
                debug("SERVER -> PROXY:\n")
                debug( response.getmessage() )

                self._PLAYreply = response
            else:
                response = Message(None, self._PLAYreply)

            # Make subsitutions and send
            cseqVal = msg.getheader("Cseq")
            response.setheader("Cseq", cseqVal)                            
            response.setheader("Session", sessionID)

            if sessionID != self.DUMMY_SESSIONID:
                debug("PROXY -> CLIENT:\n")
                debug( response.getmessage() )
                self.sendclientmsg( sessionID, response )

		# Move client from Forwarder._clients_paused to Forwarder._clients
		for k in self._forwarders.keys():
		    self._forwarders[k]._playlock.acquire()
		    self._forwarders[k]._playsessionid.append(sessionID)
		    self._forwarders[k]._playlock.release()
        else:
            debug("SESSION: Illegal input to Session.do_play. Session._proxysessionID is None\n")

#------------------------------- PAUSE ----------------------------------------------

    def do_pause( self, sessionID, msg ):
        # Store the PAUSE request for future use
        if self._PAUSErequest is None:
            self._PAUSErequest = Message(None, msg)
            
        if self._proxysessionID is not None:
            if self._PAUSEreply is None:
                # Make a PAUSEreply from a PLAYreply
                self._PAUSEreply = Message(None, self._PLAYreply)
                self._PAUSEreply.delheader("RTP-Info")
                
            # Make client-specific substitutions and send the PAUSEreply
            response = Message(None, self._PAUSEreply)
            cseqVal = msg.getheader("Cseq")
            response.setheader("Cseq", cseqVal)                            
            response.setheader("Session", sessionID)

            debug("PROXY -> CLIENT:\n")
            debug( response.getmessage() )

            self.sendclientmsg( sessionID, response )
                
            # Move client into Forwarder._clients
            for k in self._forwarders.keys():            
                self._forwarders[k]._pauselock.acquire()
                self._forwarders[k]._pausesessionid.append(sessionID)
                self._forwarders[k]._pauselock.release()
        else:
            debug("SESSION: Illegal input to Session.do_pause. Session._proxysessionID is None\n")

#------------------------------- REDIRECT  -------------------------------------------
    def do_redirect( self, sessionID, msg):
        # Send redirect message to client

        url = msg.getargs()[0]
        target = urlparse.urlparse(url, "rtsp")[4]
        
        if self._nextChild is None:
            self._nextChild = 0
        keylist = self._client_addrs.keys()
        keylist.sort()
        childSessionID = keylist[self._nextChild]
        self._nextChild = (self._nextChild + 1) % len(self._client_addrs)
        while childSessionID == sessionID or self._client_ports[childSessionID] is None or self._client_types[childSessionID] != 'chaincast ver 0.0' :
            self._nextChild = (self._nextChild + 1) % len(self._client_addrs)
            childSessionID = keylist[self._nextChild]
        
        targetChild = "rtsp://" + self._client_addrs[sessionID]
        targetChild = targetChild + ":" + self._client_ports[sessionID]
        targetChild = targetChild + "/getstream?" 

        targetChild = targetChild+"rtsp://" + self._client_addrs[childSessionID]
        targetChild = targetChild + ":" + self._client_ports[childSessionID]
        targetChild = targetChild + "/getstream?" + target
        
        redirMsgBuf = "REDIRECT " + url + " RTSP/1.0\r\n"
        redirMsgBuf = redirMsgBuf + "CSeq: 1\r\n"
        redirMsgBuf = redirMsgBuf + "Location: " + targetChild + "\r\n\r\n"
        redirectMsg = Message(None, None, redirMsgBuf)
        
	debug("PROXY -> CLIENT:\n")
	debug(redirectMsg.getmessage())
        self.sendclientmsg(sessionID, redirectMsg)
        
        # Delete this client from the session
        self.delclient(sessionID)

#------------------------------  DO_REACT_REDIRECT ------------------------------------
    def do_react_redirect( self, msg ):
        # sanity check
        if self._DESCRIBErequest is None:
            debug("Session: REDIRECT recd when self._DESCRIBErequest is None. Ignoring")
            return

        # get new server info
        newServer = msg.getheader('Location')
        args = self._DESCRIBErequest.getargs()
        args[0] = newServer
        self._DESCRIBErequest.setargs(args)

        if self._serverconn is not None:
            # Do cleanup
            self._serverconn.close()
        self._serverconn = None
        self._serveraddr = None
	self._serverport = None

        self._proxycseqval = 0
        self._proxysessionID = None
            
        # Set up connection with new server
        self.dispatch( self.DUMMY_SESSIONID, self._DESCRIBErequest )
        for k in self._SETUPrequests.keys():
            self.dispatch( self.DUMMY_SESSIONID, self._SETUPrequests[k] )

        if self._PLAYrequest is not None:
            self.dispatch( self.DUMMY_SESSIONID, self._PLAYrequest )
       
#------------------------------ TEARDOWN -----------------------------------------------
  
    def do_teardown( self, sessionID, msg):
        if self._proxysessionID is not None:

            # First store the TEARDOWN request for future use
            if self._TEARDOWNrequest is None:            
                self._TEARDOWNrequest = Message(None, msg)
            
            if self._TEARDOWNreply is None:
                # Make a TEARDOWNreply from a PLAYreply
                self._TEARDOWNreply = Message(None, self._PLAYreply)
                self._TEARDOWNreply.delheader("RTP-Info")
                if self._TEARDOWNreply.getheader("Connection") is not None:
                    self._TEARDOWNreply.setheader("Connection", "Close")
                else:
                    self._TEARDOWNreply.addheader("Connection", "Close")
            # Make client-specific substitutions and send the TEARDOWNreply
            response = Message(None, self._TEARDOWNreply)
            cseqVal = msg.getheader("Cseq")
            response.setheader("Cseq", cseqVal)                            
            response.setheader("Session", sessionID)

            debug("PROXY -> CLIENT:\n")
            debug( response.getmessage() )

            self.sendclientmsg( sessionID, response )
                
            # Remove client into Forwarder._clients, Forwarder._clients_paused
            for k in self._forwarders.keys():
                self._forwarders[k]._dellock.acquire()
                self._forwarders[k]._delsessionid.append(sessionID)
                self._forwarders[k]._dellock.release()

            # Remove client from internal data structures
            self.delclient(sessionID)
            
            if len(self._client_conns) == 0:
                # Close the upstream feed

                # Make substitutions
                self._proxycseqval = self._proxycseqval + 1
                self._TEARDOWNrequest.setheader("cseq", self._proxycseqval)
                self._TEARDOWNrequest.setheader("session", self._proxysessionID)

                # Get URL of source
                args = self._TEARDOWNrequest.getargs()
                url = args[0]
                parsed_url = urlparse.urlparse( url, "rtsp" )
                site = parsed_url[1]
                target = parsed_url[4]
                args[0] = target

                self._TEARDOWNrequest.setargs(args)
                # Send message
                self.sendservermsg( self._TEARDOWNrequest )

                # Tell parent Manager.py to delete myself from its data structures
                self._manager.delsession( target )
                
                # Close all forwarders
                for k in self._forwarders.keys():
                    del self._forwarders[k]

                # TODO: Send redirect messages to all clients
                
                debug( "SESSION: Killing myself" )
                self._amDead = 1
        else:
            debug("SESSION: Illegal input to Session.do_teardown. Session._proxysessionID is None\n")
        
#------------------------------- Run -----------------------------------------------
    def run( self ):
        #try:
            then = time.time()
	    debug("%s" % then)
            while 1:
                # Poll for new clients and call addclient() if they come
                if self._addlock.acquire(0):
                    if len(self._newclientconn) > 0:
                        debug("SESSION: Found new client info")
                        conn = self._newclientconn.pop(0)
                        addr = self._newclientaddr.pop(0)
                        cport = self._newclientport.pop(0)
                        type = self._newclienttype.pop(0)
                        msg = self._newclientmsg.pop(0)
                        sessionID = self.addclient( conn, addr, cport, type )

                        debug("SESSION: Dispatching msg")
                        self.dispatch( sessionID, msg )
                    self._addlock.release()



                # now = time.time()
		# debug("%s" % now)
		# if (now - then > 15.0):
                    #### Test code for sending REDIRECT
		    # debug("******* Sending REDIRECT ********\r\n")
                    # redirMsgBuf = "REDIRECT rtsp://fable16.stanford.edu:5982/getstream?rtsp://hake.stanford.edu:5982/broadcast RTSP/1.0\r\n"
                    # redirMsgBuf = redirMsgBuf + "CSeq: 5\r\n"
                    # redirMsgBuf = redirMsgBuf + "Location: rtsp://asli.stanford.edu:5982/getstream?rtsp://hake.stanford.edu:5982/broadcast\r\n\r\n"
                    # redirectMsg = Message(None, None, redirMsgBuf)
                    # debug("PROXY -> CLIENT:\n")
                    # debug(redirectMsg.getmessage())
		    # for k in self._client_conns.keys():
		        # self.sendclientmsg(k, redirectMsg)
		    # return


                # Else, loop across all _client_conns, and see if they have messages
                if len(self._client_conns) > 0:
                    l = []
                    for k in self._client_conns.keys():
                        l.append(self._client_conns[k])
                    readylist, _, _ = select(l, [], [], 0)
                    for k in self._client_conns.keys():
                        if self._client_conns[k] in readylist:                        
                            msg = Message( self._client_conns[k] )
                            self.dispatch( k, msg )

                # Now poll the server socket to check for any REDIRECT message
                if self._serverconn is not None:
                    readylist, _, _ = select([self._serverconn], [], [], 0)
                    if self._serverconn in readylist:
                        msg = Message( self._serverconn )

                        debug("SERVER -> PROXY unsolicited")
                        debug( msg.getmessage() )
			print "*********** REDIRECT received ************"
			print msg.getmessage()

                        self.do_react_redirect( msg )

                # See if we have to die
                if self._amDead:
                    return
                # Again
                if self._killlock.acquire(0):
                    if self._die:
                        self.do_suicide()
                        self._amDead = 1
                        return
                    self._killlock.release()

######################## SUICIDE #########################33
    def do_suicide( self ):
        redirMsgBuf = "REDIRECT " + self._url + " RTSP/1.0\r\n"
        print
        print "******** Sending redirects *********"
        print
        for k in self._client_addrs.keys():            
	    locationBuf = "Location: rtsp://" + self._client_addrs[k]
	    locationBuf = locationBuf + ":" + self._client_ports[k]
	    locationBuf = locationBuf + "/str?"
	    locationBuf = locationBuf + self._target + "\r\n"

	    msgBuf = redirMsgBuf + "CSeq: 5\r\n" + locationBuf + "\r\n"
	    redirectMsg = Message(None, None, msgBuf)
            self.sendclientmsg(k, redirectMsg)
            print
            print msgBuf
        self._serverconn.close()
        

